from cached_property import cached_property
import pandas as pd

from .uniformat import UNIBaseParser
from .misc import open_filename


class TwoWireParser(UNIBaseParser):
    def __init__(self, file):
        """
        Parameters
        ----------
        file : str
        """
        super(TwoWireParser, self).__init__(file=file)

    @cached_property
    def interval(self):
        """
        Returns
        -------
        str
        """
        _format = self.get_property(key='Format')
        # looks like ['MMR', 'Interval: 5 min']
        # so we want the second element, last part, and without the space
        return _format[1].split(': ')[1].replace(' ', '')

    def _date_parser(self, date_str, time_str):
        """
        Parameters
        ----------
        date_str : str
        time_str : str

        Returns
        -------
        pd.Timestamp
        """
        parsed = pd.Timestamp.strptime(date_str + time_str, "%d%m%Y%H:%M")
        datetime = parsed.tz_localize(self.timezone)

        return datetime

    def get_timeseries_frame(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')

    def get_metadata_frame(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')

    def _parse_dataframe(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')


class TwoWireMMRParser(TwoWireParser):
    def __init__(self, file):
        """
        Parameters
        ----------
        file : str
        """
        super(TwoWireMMRParser, self).__init__(file=file)
        self.is_long_format = False

    def _parse_dataframe(self):
        """
        Returns
        -------
        pd.DataFrame
        """
        def pandas_read(parse_dates, column_names):
            with open_filename(filename=self.file, mode='r') as f:
                _df = pd.read_csv(
                    f,
                    # because C parse does not support skipfooter:
                    engine='python',
                    header=None,
                    skiprows=self.body_start_line,
                    skipfooter=len(self.raw) - self.body_end_line - 1,
                    sep=";",
                    decimal=",",
                    true_values=['yes'],
                    false_values=['no'],
                    parse_dates=parse_dates,
                    date_parser=self._date_parser
                )
            _df.rename(columns=column_names, inplace=True)
            _df.set_index('Name', inplace=True)
            return _df

        try:
            df = pandas_read(
                parse_dates=[[5, 6], [8, 9]],
                column_names={
                    0: 'Name',
                    1: 'Type',
                    2: 'Tariff',
                    3: 'Cumulative',
                    4: 'Unit',
                    '5_6': 'Start',
                    '8_9': 'End'
                }
            )
        except ValueError:
            df = pandas_read(
                parse_dates=[[6, 7], [9, 10]],
                column_names={
                    0: 'Ean',
                    1: 'Name',
                    2: 'Type',
                    3: 'Tariff',
                    4: 'Cumulative',
                    5: 'Unit',
                    '6_7': 'Start',
                    '9_10': 'End'
                }
            )
            self.is_long_format = True

        return df

    def get_timeseries_frame(self, allow_duplicate_names=True):
        """
        Parameters
        ----------
        allow_duplicate_names : bool
            default True
            if False, only the first occurence of a name is returned
            
        Returns
        -------
        pd.DataFrame
        """
        df = self.get_dataframe()

        start = df.Start.iloc[0]
        end = df.End.iloc[0]
        index = pd.date_range(start=start, end=end, freq=self.interval)

        if not self.is_long_format:
            vals = df[df.columns[6:]]
        else:
            vals = df[df.columns[7:-2]]
            
        if not allow_duplicate_names:
            vals = vals.reset_index()
            vals = vals.drop_duplicates(subset='Name')
            vals = vals.set_index('Name')

        vals = vals.T
        vals.index = index

        return vals

    def get_metadata_frame(self, allow_duplicate_names=True):
        """
        Parameters
        ----------
        allow_duplicate_names : bool
            default True
            if False, only the first occurence of a name is returned
            
        Returns
        -------
        pd.DataFrame
        """
        df = self.get_dataframe()
        if not self.is_long_format:
            meta = df[df.columns[:6]]
        else:
            meta = df[df.columns[:7]]

        if not allow_duplicate_names:
            meta = meta.reset_index()
            meta = meta.drop_duplicates(subset='Name')
            meta = meta.set_index('Name')

        return meta.T



class open_filename(object):
    """
    Context manager that opens a filename and closes it on exit, but does
    nothing for file-like objects.
    """
    def __init__(self, filename, *args, **kwargs):
        self.closing = kwargs.pop('closing', False)
        if isinstance(filename, str):
            self.fh = open(filename, *args, **kwargs)
            self.closing = True
        else:
            self.fh = filename

    def __enter__(self):
        self.fh.seek(0)
        return self.fh

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.closing:
            self.fh.close()
        else:
            self.fh.seek(0)

        return False



import csv
from cached_property import cached_property
import pytz
import datetime as dt

from .misc import open_filename


class UNIBaseParser:
    def __init__(self, file):
        """
        Parameters
        ----------
        file : str
        """
        self.file = file
        self.body_start_line = None
        self.body_end_line = None

        with open_filename(filename=file, mode='r') as f:
            self.raw = list(csv.reader(f, delimiter=";"))

        self.dict = self._parse_properties(raw=self.raw)

        self.df = None

    @property
    def properties(self):
        """
        Returns
        -------
        set
        """
        return set(self.dict.keys())

    def get_property(self, key):
        """
        Parameters
        ----------
        key : str

        Returns
        -------
        str | list(str) | None
        """
        return self.dict.get(key)

    @cached_property
    def timezone(self):
        """
        Get the timezone offset

        Returns
        -------
        pytz.FixedOffset
        """
        tz_string = self.get_property(key="Time zone")
        sign = tz_string[0]
        hours = int(tz_string[1:3])
        minutes = int(tz_string[3:5])

        offset = dt.timedelta(hours=hours, minutes=minutes)
        offset_minutes = offset.seconds / 60
        offset_signed = float(sign + str(offset_minutes))
        return pytz.FixedOffset(offset_signed)

    def _parse_properties(self, raw):
        """
        Transform raw lines into dict

        Parameters
        ----------
        raw : list(list(str))

        Returns
        -------
        dict
        """
        d = {}

        for i, line in enumerate(raw):
            key = line[0]

            # all keys start and end with square braces
            if not key.startswith('['):
                continue  # skip this line
            else:
                key = key.strip("[]")

            if key == "Body Start":
                self.body_start_line = i + 1
                continue
            elif key == "Body End":
                self.body_end_line = i - 1
                continue

            # get everything after the key, but no empty strings
            value = [elem for elem in line[1:] if elem != '']

            if len(value) == 1:
                value = value[0]
            elif len(value) == 0:
                continue

            d.update({key: value})

        return d

    def get_dataframe(self):
        """
        Returns
        -------
        pd.DataFrame
        """
        if self.df is not None:
            return self.df
        else:
            return self._parse_dataframe()

    def get_timeseries_frame(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')

    def get_metadata_frame(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')

    def _parse_dataframe(self):
        """
        Returns
        -------
        pd.DataFrame

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError('Method needs to be implemented by subclass')



# -*- coding: utf-8 -*-

"""
A setuptools based setup module for smappy.

Adapted from
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version from the source code
with open(path.join(here, 'ediel', 'ediel.py'), encoding='utf-8') as f:
    lines = f.readlines()
    for l in lines:
        if l.startswith('__version__'):
            __version__ = l.split('"')[1] # take the part after the first "

setup(
    name='ediel',
    version=__version__,
    description='A Python parser for EDIEL and UNIFORMAT Energy Data',
    long_description=long_description,
    url='https://github.com/energyid/ediel',
    author='EnergieID.be',
    author_email='jan@energieid.be',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='ediel uniformat energy parser',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #py_modules=[""],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed.
    install_requires=['pandas'],

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # Note: for creating the source distribution, they had to be included in the
    # MANIFEST.in as well.
    package_data={
        'ediel': ['LICENSE', 'README.md'],
    },
)
